import numpy as np
from ._version import __version__

__all__ = ["flux_to_mag", "flux_err_to_mag_err", "print_version", "elk_say"]


_elk_name = r"""      _ _
     | | |
  ___| | | __
 / _ \ | |/ /
|  __/ |   <
 \___|_|_|\_\ """

_elk_say_base = r"""     /|        |\
  `__\\        //__'
     ||        ||
   \__`\      /'__/
     `_\\    //_'
     _.:------:._
     \_: O..O :_/
       |      |
       \.----./
        \ oo /
         \__/"""

""" ANSI color codes """
RED = "\033[0;31m"
GREEN = "\033[0;32m"
YELLOW = "\033[1;33m"
END = "\033[0m"


def flux_to_mag(flux):
    """Converts flux to TESS Magnitudes

    Parameters
    ----------
    flux : :class:`~numpy.ndarray` or `float`
        flux in terms of ergs/s

    Returns
    -------
    mag : :class:`~numpy.ndarray` or `float`
        TESS Magnitude centered in the I band
    """
    m1 = 10     # calibration mag
    f1 = 15000  # Calibration Flux corresponding to the calibration mag
    mag = 2.5 * np.log10(f1 / flux) + m1
    return mag


def flux_err_to_mag_err(flux, flux_err):
    """Converts TESS Flux Errors to Magnitude errors

    Parameters
    ----------
    flux : class: `~numpy.ndarray` or `float`
        flux in terms of ergs/s
    flux_err : class: `~numpy.ndarray` or `float`
       flux error in terms of ergs/s

    Returns
    -------
   mag_err : class: `~numpy.ndarray` or `float`
        TESS Magnittude errors
    """
    d_mag_d_flux = -2.5 / (flux * np.log(10))
    m_err_squared = abs(d_mag_d_flux)**2 * flux_err**2
    return np.sqrt(m_err_squared)

def print_warning(*args):
    """Print a warning in yellow, prefixed by WARNING"""
    print(f"{YELLOW}WARNING:{END}", *args)


def print_failure(*args):
    """Print something in red"""
    print(RED, end="")
    print(*args, END)


def print_success(*args):
    """Print something in green"""
    print(GREEN, end="")
    print(*args, END)


def print_version():
    """Print out the current version of `elk` in a fancy way"""
    print(f'{_elk_name}version {__version__}')


def _pad_space(s, n):
    """Pad a string with some spaces to the left"""
    return n * ' ' + s


def elk_say(message):
    """Get the elk mascot to say something

    Parameters
    ----------
    message : `str`
        The message that you want to print
    """
    dashes = '-' * len(message)
    print(dashes)
    print(message)
    print(dashes)

    halfway = len(message) // 2
    for i in range(2):
        print(_pad_space('\\', halfway + i))

    for line in _elk_say_base.split('\n'):
        print(_pad_space(line, halfway))


import numpy as np
import astropy.units as u
from astropy.io import fits
from astropy.table import Table, vstack, Column
from astropy.timeseries import LombScargle
import lightkurve as lk
from tqdm import tqdm
import matplotlib.pyplot as plt

import os
import imageio
from astroquery.simbad import Simbad
from IPython.display import HTML
import astropy.coordinates as coord

from .utils import flux_to_mag, flux_err_to_mag_err
import elk.plot as elkplot
import elk.stats as elkstats


__all__ = ["BasicLightcurve", "TESSCutLightcurve"]


TESS_RESOLUTION = 21 * u.arcsec / u.pixel


class BasicLightcurve():
    def __init__(self, time=None, flux=None, flux_err=None, sector=None, fits_path=None, hdu_index=None,
                 periodogram_freqs=np.arange(0.04, 11, 0.01)):
        """A basic lightcurve class for calculating statistics of corrected lightcurves and plotting them.

        This class can either be instantiated manually using time, flux and flux_err values, or loaded in from
        a fits file that has previously been created.

        Parameters
        ----------
        time : :class:`~numpy.ndarray`, optional
            Time of observations, by default None
        flux : :class:`~numpy.ndarray`, optional
            Flux from each observations, by default None
        flux_err : :class:`~numpy.ndarray`, optional
            Errors on the flux, by default None
        sector : `int`, optional
            TESS sector in which observations were taken, by default None
        fits_path : `str`, optional
            Path to a fits file containing the lightcurve, by default None
        hdu_index : `int`, optional
            Index of the HDU containing the lightcurve, by default None
        periodogram_freqs : :class:`numpy.ndarray`, optional
            Frequencies at which to evaluate any periodograms, by default np.arange(0.04, 11, 0.01)
        """
        # check that either data or a file has been given
        has_data = time is not None and flux is not None and flux_err is not None and sector is not None
        has_file = fits_path is not None and hdu_index is not None
        assert has_data or has_file, "Must either provide data or a fits file and HDU index"

        # if the user has given us a file then load it into the class
        if has_file:
            with fits.open(fits_path) as hdul:
                hdu = hdul[hdu_index]
                self.corrected_lc = lk.LightCurve(time=hdu.data["time"] * u.day,
                                                  flux=hdu.data["flux"] * u.electron / u.s,
                                                  flux_err=hdu.data["flux_err"] * u.electron / u.s)
                self.sector = hdu.header["sector"]
        # otherwise create a lightcurve directly
        else:
            self.corrected_lc = lk.LightCurve(time=time, flux=flux, flux_err=flux_err)
            self.sector = sector

        # set up an HDU for saving lightcurves
        self.hdu = fits.BinTableHDU.from_columns(
            [fits.Column(name='time', format='D', array=self.corrected_lc.time.value),
             fits.Column(name='flux', format='D', array=self.corrected_lc.flux.value),
             fits.Column(name='flux_err', format='D', array=self.corrected_lc.flux_err.value),
             fits.Column(name='mag', format='D', array=flux_to_mag(self.corrected_lc.flux.value)),
             fits.Column(name='mag_err', format='D',
                         array=flux_err_to_mag_err(self.corrected_lc.flux.value,
                                                   self.corrected_lc.flux_err.value))]
        )
        self.hdu.header.set('sector', self.sector)

        # prep some class variables for the stats helper functions
        self.stats = {}
        self.periodogram_freqs = periodogram_freqs
        self.periodogram_peaks = None
        self.ac_time = None

    def __repr__(self):
        return f"<{self.__class__.__name__}: TESS Sector {self.sector}, {len(self.corrected_lc)} timesteps>"

    @property
    def normalized_flux(self):
        """The normalised flux of the lightcurve"""
        return self.corrected_lc.flux.value / np.median(self.corrected_lc.flux.value)

    @property
    def rms(self):
        """The root-mean-squared normalised flux of the lightcurve"""
        self.stats["rms"] = np.sqrt(np.mean(self.normalized_flux**2))
        return self.stats["rms"]

    @property
    def std(self):
        """The standard deviation of the normalised flux of the lightcurve"""
        self.stats["std"] = np.std(self.normalized_flux)
        return self.stats["std"]

    @property
    def MAD(self):
        """The median absolute deviation of the normalised flux of the lightcurve"""
        self.stats["MAD"] = elkstats.get_MAD(self.normalized_flux)
        return self.stats["MAD"]
    
    @property
    def sigmaG(self):
        """The sigmaG of the normalised flux of the lightcurve"""
        self.stats["sigmaG"] = elkstats.get_sigmaG(self.normalized_flux)
        return self.stats["sigmaG"]

    @property
    def skewness(self):
        """The skewness of the flux of the lightcurve"""
        self.stats["skewness"] = elkstats.get_skewness(self.corrected_lc.flux.value)
        return self.stats["skewness"]

    @property
    def von_neumann_ratio(self):
        """The Von Neumann ratio for the normalised flux of the lightcurve. See
        :class:`elk.stats.von_neumann_ratio` for more information"""
        self.stats["von_neumann_ratio"] = elkstats.von_neumann_ratio(self.normalized_flux)
        return self.stats["von_neumann_ratio"]

    def J_stetson(self, **kwargs):
        """Calculate the J Stetson statistic for the lightcurve

        Keyword arguments are passed directly to :class:`elk.stats.J_stetson`

        Returns
        -------
        J_Stetson
            The J Stetson statistic
        """
        mag_err = flux_err_to_mag_err(self.corrected_lc.flux.value, self.corrected_lc.flux_err.value)
        self.stats["J_Stetson"] = elkstats.J_stetson(time=self.corrected_lc.time.value,
                                                     mag=flux_to_mag(self.corrected_lc.flux.value),
                                                     mag_err=mag_err, **kwargs)
        return self.stats["J_Stetson"]

    def to_periodogram(self, frequencies=None, **kwargs):
        """Construct a periodogram using the lightcurve

        Parameters
        ----------
        frequencies : :class:`numpy.ndarray`, optional
            Frequencies at which to evaluate the periodogram, by default self.periodogram_freqs
        **kwargs : `various`
            Keyword arguments to pass to :class:`elk.stats.periodogram`

        Returns
        -------
        Same as :class:`elk.stats.periodogram`
        """
        # use default frequencies
        if frequencies is None:
            frequencies = self.periodogram_freqs

        self.periodogram, self.periodogram_percentiles,\
            lsp_stats = elkstats.periodogram(self.corrected_lc.time.value, self.corrected_lc.flux.value,
                                             self.corrected_lc.flux_err.value, frequencies=frequencies,
                                             **kwargs)
        self.periodogram_freqs = frequencies
        self.periodogram_peaks = lsp_stats["peak_freqs"]
        self.stats.update(lsp_stats)
        return self.periodogram, self.periodogram_percentiles, lsp_stats

    def to_acf(self, **kwargs):
        """Calculate the autocorrelation function for the lightcurve

        Keyword arguments are passed directly to :class:`elk.stats.autocorr`

        Returns
        -------
        Same as :class:`elk.stats.periodogram`
        """
        r = elkstats.autocorr(time=self.corrected_lc.time.value, flux=self.corrected_lc.flux.value, **kwargs)
        self.ac_time, self.acf, self.acf_percentiles, acf_stats = r[:4]
        self.stats.update(acf_stats)
        return r

    def get_stats_using_defaults(self):
        """Generate all statistics for the lightcurve **using default settings**

        NOTE: This may not be optimal for your use case, be sure to consider each setting! By default we use
        a list of frequencies generated with ``np.arange(0.04, 11, 0.01)`` for the periodogram.

        Returns
        -------
        stats : `dict`
            A dictionary of the various statistics (also stored in ``self.stats``)
        """
        (self.rms, self.std, self.MAD, self.sigmaG, self.skewness,
            self.von_neumann_ratio, self.J_stetson(),
            self.to_periodogram(frequencies=np.arange(0.04, 11, 0.01)), self.to_acf())
        return self.stats

    def get_stats_table(self, name, run_all=False):
        """Create an Astropy Table of the statistics (for aggregation with other lightcurves)

        Parameters
        ----------
        name : `str`
            Identifier for this lightcurve (e.g. name of cluster)
        run_all : `bool`, optional
            Whether to run all statistics **using default settings**
            (see ``.get_stats_using_defaults()``), by default False

        Returns
        -------
        stats_table : :class:`~astropy.table.Table`
            Table of statistics
        """
        if run_all:
            self.get_stats_using_defaults()
        table_dict = {"name": [name]}
        for k, v in self.stats.items():
            table_dict[k] = [v]
        return Table(table_dict)

    def plot(self, title="auto", fold_period=None, **kwargs):
        """Plot the lightcurve

        Parameters
        ----------
        title : `str`, optional
            Title for the plot, by default "auto" (resulting in "Lightcurve for Sector ...")
        fold_period : `float`, optional
            Period on which to fold the light curve, if None then no folding is done, by default None
        **kwargs: `various`
            Keyword arguments passed to :class:`elk.plot.plot_lightcurve`

        Returns
        -------
        fig, ax : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
            Figure and axis on which the lightcurve has been plotted
        """
        title = f'Lightcurve for Sector {self.sector}' if title == "auto" else title
        return elkplot.plot_lightcurve(self.corrected_lc.time.value, self.corrected_lc.flux.value,
                                       title=title, fold_period=fold_period, **kwargs)

    def plot_periodogram(self, frequencies=None, title="auto", **kwargs):
        """Plot the periodogram for the lightcurve

        Parameters
        ----------
        frequencies : :class:`~numpy.ndarray`
            Frequencies at which periodogram evaluated, if None then periodogram must have already been
            generated using ``.to_periodogram()``
        title : `str`, optional
            Title for the plot, by default "auto" (resulting in "Periodogram for Sector ...")
        **kwargs: `various`
            Keyword arguments passed to :class:`elk.plot.plot_periodogram`

        Returns
        -------
        fig, ax : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
            Figure and axis on which the periodogram has been plotted
        """
        # get the periodogram if it hasn't already been run
        if "peak_freqs" not in self.stats:
            self.to_periodogram(frequencies=frequencies)

        title = f'Periodogram for Sector {self.sector}' if title == "auto" else title
        return elkplot.plot_periodogram(frequencies=self.periodogram_freqs, power=self.periodogram,
                                        power_percentiles=self.periodogram_percentiles,
                                        peak_freqs=self.stats["peak_freqs"][:self.stats["n_peaks"]],
                                        fap=self.stats["FAP"], title=title, **kwargs)

    def plot_acf(self, title="auto", **kwargs):
        """Plot the autocorrelation function

        Parameters
        ----------
        title : `str`, optional
            Title for the plot, by default "auto" (resulting in "Autocorrelation function for Sector ...")
        **kwargs: `various`
            Keyword arguments passed to :class:`elk.plot.plot_acf`

        Returns
        -------
        fig, ax : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
            Figure and axis on which the autocorrelation function has been plotted
        """
        if self.ac_time is None:
            self.to_acf()

        title = f'Autocorrelation function for Sector {self.sector}' if title == "auto" else title
        return elkplot.plot_acf(time=self.ac_time, acf=self.acf, acf_percentiles=self.acf_percentiles,
                                title=title, **kwargs)

    def analysis_plot(self, name=None, run_all=False, show=True):
        """Plot a 3-panel plot of the lightcurve for quick analysis

        Plot includes the lightcurve, periodogram and autocorrelation function.

        Parameters
        ----------
        name : `str`
            Identifier for this lightcurve (e.g. name of cluster)
        run_all : `bool`, optional
            Whether to run all statistics **using default settings** (see ``.get_stats_using_defaults()``),
            by default False
        show : `bool`, optional
            Whether to immediately show the plot, by default True

        Returns
        -------
        fig, axes : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
            Figure and axes on which the analysis has been plotted
        """
        # run everything if the user wants to
        if run_all:
            self.get_stats_using_defaults()

        fig, axes = plt.subplots(1, 3, figsize=(30, 5))

        name = "Lightcurve Analysis Plot" if name is None else name
        plt.suptitle(f'{name} (Sector {self.sector})', fontsize="xx-large")

        self.plot(fig=fig, ax=axes[0], show=False, title="Lightcurve")
        self.plot_periodogram(fig=fig, ax=axes[1], show=False, title="Periodogram")
        self.plot_acf(fig=fig, ax=axes[2], show=show, title="Autocorrelation function")

        return fig, axes


class TESSCutLightcurve(BasicLightcurve):
    def __init__(self, radius, lk_search_result=None, tpfs=None,
                 cutout_size=99, percentile=80, n_pca=6, spline_knots=5,
                 periodogram_freqs=np.arange(0.04, 11, 0.01), save_pixel_periodograms=True,
                 progress_bar=False):
        """A lightcurve constructed from a TESSCut search with various correction functionalities

        Parameters
        ----------
        radius : `float`
            Radius of the cluster in degrees.
        lk_search_result : :class:`lightkurve.SearchResult`, optional
            Search result from a LightKurve tesscut call, by default None
        tpfs : :class:`lightkurve.TessTargetPixelFile`, optional
            Target pixel files, by default None
        cutout_size : `int`, optional
            Cutout size for the TESSCut call, by default 99
        percentile : `int`, optional
            Which percentile to use in the upper limit calculation, by default 80
        n_pca : `int`, optional
            Number of principle components to use in the DesignMatrix, by default 6
        spline_knots: `int`, optional
            Number of knots to include in the spline corrector, by default 5. If None, spline corrector
            will not be used
        periodogram_freqs : :class:`numpy.ndarray`, optional
            Frequencies at which to evaluate any periodograms, by default np.arange(0.04, 11, 0.01)
        progress_bar : `bool`, optional
            Whether to show a progress bar of pixel correction, by default False
        """

        assert lk_search_result is not None or tpfs is not None, "Must supply either a search result or tpfs"

        # convert radius to degrees if it has units
        if hasattr(radius, 'unit'):
            radius = radius.to(u.deg).value

        self.radius = radius
        self.lk_search_results = lk_search_result
        self._tpfs = tpfs
        self.cutout_size = cutout_size
        self.percentile = percentile
        self.n_pca = n_pca
        self.spline_knots = spline_knots
        self.save_pixel_periodograms = save_pixel_periodograms
        self.progress_bar = progress_bar

        self.periodogram_freqs = periodogram_freqs
        self.pixel_periodograms = None if not self.save_pixel_periodograms else [None for _ in
                                                                                 range(self.cutout_size**2)]

        # defaults for the cached variables
        self._quality_tpfs = None
        self._basic_lc = None
        self._quality_lc = None
        self._uncorrected_lc = None

        # prep some class variables for the stats helper functions
        self.stats = {}
        self.ac_time = None

    @property
    def tpfs(self):
        """All target pixel files for the lightcurve"""
        if self._tpfs is None:
            self._tpfs = self.lk_search_results.download(cutout_size=(self.cutout_size, self.cutout_size))
        return self._tpfs

    @property
    def sector(self):
        """TESS sector in which observations were taken"""
        return self.tpfs.sector

    @property
    def quality_tpfs(self):
        """Target pixel files that have a quality flag of 0 and a positive flux_err"""
        if self._quality_tpfs is None:
            self._quality_tpfs = self.tpfs[(self.basic_lc.quality == 0) & (self.basic_lc.flux_err > 0)]
        return self._quality_tpfs

    @property
    def basic_lc(self):
        """Lightcurve constructed using **all** target pixel files"""
        if self._basic_lc is None:
            self._basic_lc = self.tpfs.to_lightcurve()
        return self._basic_lc

    @property
    def quality_lc(self):
        """Lightcurve constructed using only quality target pixel files"""
        if self._quality_lc is None:
            self._quality_lc = self.quality_tpfs.to_lightcurve()
        return self._quality_lc

    @property
    def uncorrected_lc(self):
        """Lightcurve constructed using only quality target pixel files and a circle aperture mask"""
        if self._uncorrected_lc is None:
            self.star_mask = self.circle_aperture()
            self._uncorrected_lc = self.quality_tpfs.to_lightcurve(aperture_mask=self.star_mask)
        return self._uncorrected_lc

    def fails_quality_test(self):
        """Test whether this lightcurve has (1) any quality TPFs, (2) that observations are not within ~0.28
        degrees of the edge of the detector (such that NaN values would appear in the cutout) and (3) that it
        isn't part of TESS Sector 1, which has an unremovable systematic.

        Returns
        -------
        flag : `bool`
            Flag of whether the test was passed
        """
        # check that there is at least one good quality TPF
        if not np.any((self.basic_lc.quality == 0) & (self.basic_lc.flux_err > 0)):
            return True

        min_flux = np.min(self.quality_tpfs[0].flux.value)
        min_not_nan = ~np.isnan(min_flux)
        # Also making sure the Sector isn't the one with the Systematic
        not_sector_one = self.tpfs.sector != 1
        min_flux_greater_one = min_flux > 1
        return ~(min_not_nan & not_sector_one & min_flux_greater_one)

    def circle_aperture(self):
        """Generate a circular aperture mask based on the radius and cutout_size of this lightcurve

        Returns
        -------
        mask : :class:`numpy.ndarray`
            Aperture mask
        """
        # convert the radius to pixels based on the TESS resolution
        radius_in_pixels = (self.radius * u.deg / TESS_RESOLUTION).to(u.pixel).value

        # mask the grid of pixels based on this radius
        pix_x, pix_y = np.meshgrid(np.arange(self.cutout_size), np.arange(self.cutout_size))
        return (pix_x - self.cutout_size // 2)**2 + (pix_y - self.cutout_size // 2)**2 < radius_in_pixels**2

    def correct_lc(self):
        """Correct the lightcurve using the method described in Wainer+2023"""
        # Time average of the pixels in the TPF:
        max_frame = self.quality_tpfs.flux.value.max(axis=0)

        # This renormalizes any columns which are bright because of straps on the detector
        max_frame -= np.median(max_frame, axis=0)

        # This aperture is any "faint" pixels:
        bkg_aper = max_frame < np.percentile(max_frame, self.percentile)

        # The average light curve of the faint pixels is a good estimate of the scattered light
        scattered_light = self.quality_tpfs.flux.value[:, bkg_aper].mean(axis=1)

        # We can use our background aperture to create pixel time series and then take Principal
        # Components of the data using Singular Value Decomposition. This gives us the "top" trends
        # that are present in the background data
        pca_dm = lk.DesignMatrix(self.quality_tpfs.flux.value[:, bkg_aper], name='PCA').pca(self.n_pca)

        # Here we are going to set the priors for the PCA to be located around the
        # flux values of the uncorrected LC
        pca_dm.prior_mu = np.array([np.median(self.uncorrected_lc.flux.value) for _ in range(self.n_pca)])
        pca_dm.prior_sigma = np.array([(np.percentile(self.uncorrected_lc.flux.value, 84)
                                         - np.percentile(self.uncorrected_lc.flux.value, 16))
                                        for _ in range(self.n_pca)])

        # The TESS mission pipeline provides co-trending basis vectors (CBVs) which capture common trends
        # in the dataset. We can use these to de-trend out pixel level data. The mission provides
        # MultiScale CBVs, which are at different time scales. In this case, we don't want to use the long
        # scale CBVs, because this may fit out real astrophysical variability. Instead we will use the
        # medium and short time scale CBVs.
        cbvs_1 = lk.correctors.cbvcorrector.load_tess_cbvs(sector=self.quality_tpfs.sector,
                                                           camera=self.quality_tpfs.camera,
                                                           ccd=self.quality_tpfs.ccd,
                                                           cbv_type='MultiScale',
                                                           band=2).interpolate(self.quality_lc,
                                                                               extrapolate=True)
        cbvs_2 = lk.correctors.cbvcorrector.load_tess_cbvs(sector=self.quality_tpfs.sector,
                                                           camera=self.quality_tpfs.camera,
                                                           ccd=self.quality_tpfs.ccd,
                                                           cbv_type='MultiScale',
                                                           band=3).interpolate(self.quality_lc,
                                                                               extrapolate=True)

        cbv_dm1 = cbvs_1.to_designmatrix(cbv_indices=np.arange(1, 8))
        cbv_dm2 = cbvs_2.to_designmatrix(cbv_indices=np.arange(1, 8))

        # This combines the different timescale CBVs into a single `designmatrix` object
        cbv_dm_use = lk.DesignMatrixCollection([cbv_dm1, cbv_dm2]).to_designmatrix()

        # We can make a simple basis-spline (b-spline) model for astrophysical variability. This will be a
        # flexible, smooth model. The number of knots is important, we want to only correct for very long
        # term variability that looks like systematics, so here we have 5 knots, the smaller the better
        spline_dm1 = lk.designmatrix.create_spline_matrix(self.quality_tpfs.time.value, 
                                                            n_knots=self.spline_knots)
        self.dm = lk.DesignMatrixCollection([pca_dm, cbv_dm_use, spline_dm1])

        full_model, systematics_model, self.full_model_normalized = np.ones((3, *self.quality_tpfs.shape))
        if self.progress_bar:
            for i, j in tqdm([(i, j) for i in range(self.cutout_size) for j in range(self.cutout_size)]):
                systematics_model[:, i, j],\
                    full_model[:, i, j], self.full_model_normalized[:, i, j] = self.correct_pixel(i, j)
        else:
            for i, j in [(i, j) for i in range(self.cutout_size) for j in range(self.cutout_size)]:
                systematics_model[:, i, j],\
                    full_model[:, i, j], self.full_model_normalized[:, i, j] = self.correct_pixel(i, j)

        # Calculate Lightcurves
        self.corrected_lc = (self.quality_tpfs - full_model).to_lightcurve(aperture_mask=self.star_mask)
            
        self.systematics_corrected_lc = (self.quality_tpfs - systematics_model).to_lightcurve(aperture_mask=self.star_mask)

        self.hdu = fits.BinTableHDU.from_columns(
            [fits.Column(name='time', format='D', array=self.corrected_lc.time.value),
             fits.Column(name='flux', format='D', array=self.corrected_lc.flux.value),
             fits.Column(name='flux_err', format='D', array=self.corrected_lc.flux_err.value),
             fits.Column(name='mag', format='D', array=flux_to_mag(self.corrected_lc.flux.value)),
             fits.Column(name='mag_err', format='D',
                         array=flux_err_to_mag_err(self.corrected_lc.flux.value,
                                                   self.corrected_lc.flux_err.value))]
        )
        self.hdu.header.set('sector', self.sector)

    def correct_pixel(self, i, j):
        """Correct an individual pixel of the lightcurve

        Parameters
        ----------
        i, j : `int`
            Indices for the pixel

        Returns
        -------
        systematics_model : :class:`numpy.ndarray`
            A model for the systematics in the pixel
        full_model : :class:`numpy.ndarray`
            The full model for the pixel lightcurve
        full_model_normalized : :class:`numpy.ndarray`
            The normalised model for the pixel lightcurve
        """
        # create a lightcurve for just this pixel
        pixel_lightcurve = lk.LightCurve(time=self.quality_tpfs.time.value,
                                         flux=self.quality_tpfs.flux.value[:, i, j],
                                         flux_err=self.quality_tpfs.flux_err.value[:, i, j])

        # create a regression corrector based on the design matrix
        r1 = lk.RegressionCorrector(pixel_lightcurve)

        # correct the pixel lightcurve by our design matrix
        corrected_lc = r1.correct(self.dm)

        if self.save_pixel_periodograms:
            lsp = LombScargle(t=corrected_lc["time"], y=corrected_lc["flux"], dy=corrected_lc["flux_err"])
            self.pixel_periodograms[i * self.cutout_size + j] = lsp.power(self.periodogram_freqs / u.day)

        # extract just the systematics components
        systematics_model = (r1.diagnostic_lightcurves['PCA'].flux.value
                             + r1.diagnostic_lightcurves['CBVs'].flux.value)

        # normalise the model for scattered light test
        full_model_normalized = systematics_model + r1.diagnostic_lightcurves['spline'].flux.value

        # add all the components and adjust by the mean to avoid being centered around 0
        full_model = full_model_normalized - r1.diagnostic_lightcurves['spline'].flux.value.mean()
        
        del r1

        return systematics_model, full_model, full_model_normalized

    def diagnose_lc_periodogram(self, output_path, freq_bins='auto', identifier='', query_simbad=True):
        """Create gif showing pixels that contribute power to periodogram for different frequency ranges

        The GIF has 3 panels, the first shows the overall TPFs and aperture, the second shows the maximum
        power in each pixel for this frequency bin (and is annotated with the frequency/range used for this
        frame) and the last panel shows the overall ensemble light curve periodogram with the frame's
        frequency range highlighted.

        NOTE: `self.correct_lc` must have already been run and `self.save_pixel_periodograms` must be true.

        Parameters
        ----------
        output_path : `str`
            Path to a folder in which to output the gif and frames
        freq_bins : `str` or `int` or :class:`~numpy.ndarray`, optional
            Frequency bins to use for the GIF. Either 'auto' to create a frame for each peak in the
            periodogram, an integer to use log-spaced bins in the periodogram range or an array of bin edges,
            by default 'auto'
        identifier : `str`, optional
            An identifier for this target to put in the title (e.g. the cluster name), by default ''
        """
        # ensure the necessary data is available to run this
        assert self.save_pixel_periodograms, ("Pixel periodograms not available. Set "
                                              "`self.save_pixel_periodograms=True` and re-run "
                                              "`self.correct_lc`")

        # mask the pixel powers to only be for pixels in the aperture
        assert hasattr(self, "star_mask"), "No aperture mask found - have you run `self.correct_lc`?"
        aperture_powers = np.asarray(self.pixel_periodograms)[self.star_mask.flatten()]
        assert len(aperture_powers) > 0, "No pixel periodograms found - did you run `self.correct_lc`?"

        # ensure output folder and subfolder exists
        assert os.path.exists(output_path), f"No folder found at output_path: {output_path}"
        if not os.path.exists(os.path.join(output_path, 'diagnostics')):
            os.mkdir(os.path.join(output_path, 'diagnostics'))

        if query_simbad:
            simbad_results = None

        # convert input into bin bounds
        if isinstance(freq_bins, str):
            assert freq_bins == "auto", "`freq_bins` can only be a str if it is equal to 'auto'"
            self.to_periodogram()
            bounds = list(zip(self.stats["peak_left_edge"], self.stats["peak_freqs"],
                              self.stats["peak_right_edge"]))
        else:
            if isinstance(freq_bins, int):
                freq_bins = np.logspace(min(self.periodogram_freqs), max(self.periodogram_freqs), freq_bins)
            bounds = list(zip(freq_bins[:-1], 0.5 *(freq_bins[:-1] + freq_bins[1:]), freq_bins[1:]))

        # create a separate frame for each frequency bin
        i = 0
        for lower, center, upper in bounds:
            if lower > max(self.periodogram_freqs):
                continue
            # start a three panel figure
            fig, axes = plt.subplots(1, 3, figsize=(18, 4))

            # plot the entire target pixel file in the first axis
            self.quality_tpfs.plot(frame=len(self.quality_tpfs) // 2, ax=axes[0])

            # create a circle around the aperture
            x_range = axes[0].get_xlim()[1] - axes[0].get_xlim()[0]
            y_range = axes[0].get_ylim()[1] - axes[0].get_ylim()[0]
            circle = plt.Circle(xy=(axes[0].get_xlim()[0] + x_range / 2, axes[0].get_ylim()[0] + y_range / 2),
                                radius=(self.radius * u.deg / TESS_RESOLUTION).to(u.pixel).value,
                                edgecolor="red", facecolor="none", linewidth=2)
            axes[0].add_artist(circle)
            axes[0].set_title(f'{identifier} ({self.quality_tpfs[0].ra}, {self.quality_tpfs[0].dec})')

            # get the indices of the aperture pixels and plot a marker on each pixel that matches
            pixel_inds = np.argwhere(self.star_mask)
            axes[0].scatter(axes[0].get_xlim()[0] + pixel_inds[:, 1] + 0.5,
                            axes[0].get_ylim()[0] + pixel_inds[:, 0] + 0.5,
                            c='r', s=1, alpha=0.5)

            # create a mask for the frequency range
            frequency_mask = (self.periodogram_freqs >= lower) & (self.periodogram_freqs < upper)

            # get power in each pixel that is within the aperture and for the given frequency range
            pixel_powers = aperture_powers[:, frequency_mask]

            # get the max power in each pixel
            pixel_max_power = np.zeros([self.cutout_size, self.cutout_size], dtype='float64')
            pixel_max_power[self.star_mask] = np.max(pixel_powers, axis=1)

            if query_simbad:
                query_pixels = np.argwhere(pixel_max_power > 0.9 * np.max(pixel_max_power))

                world_values = np.array(self.quality_tpfs.wcs.pixel_to_world_values(np.fliplr(query_pixels)))
                ra, dec = world_values[:, 0], world_values[:, 1]

                # create custom simbad that includes variable star columns
                var_Simbad = Simbad()
                var_Simbad.add_votable_fields('v*', 'otype', 'flux(V)')

                # query SIMBAD for a region that fully encloses the pixel
                pixel_radius = (TESS_RESOLUTION * (1 * u.pixel)).to(u.deg) * np.sqrt(2)
                query_result = var_Simbad.query_region(coord.SkyCoord(ra=ra, dec=dec,
                                                       unit=(u.deg, u.deg), frame='icrs'),
                                                       radius=pixel_radius)

                query_result=query_result['MAIN_ID', 'RA', 'DEC', 'V__vartyp', 'V__Vmax', 'V__R_Vmax', 
                                          'V__magtyp', 'V__UpVmin', 'V__Vmin', 'V__R_Vmin', 'V__UpPeriod', 
                                          'V__period', 'V__R_period', 'OTYPE', 'FLUX_V']
                
                if query_result is not None:
                    # only selecting the columns we want
                    query_result = query_result['MAIN_ID', 'RA', 'DEC', 'V__vartyp', 'V__Vmax', 'V__R_Vmax',
                                                'V__magtyp', 'V__UpVmin', 'V__Vmin', 'V__R_Vmin',
                                                'V__UpPeriod', 'V__period', 'V__R_period', 'OTYPE', 'FLUX_V']
                    
                    # add columns indicating which peak these are associated with
                    query_result.add_column(Column([round(center, 3)]), name='peak_freq')
                    query_result.add_column(Column([round(lower, 3)]), name='peak_lower')
                    query_result.add_column(Column([round(upper, 3)]), name='peak_upper')

                    if simbad_results is None:
                        # if this is the first query then just save
                        simbad_results = query_result
                    else:
                        # otherwise concatenate with previous queries
                        simbad_results = vstack([simbad_results, query_result],
                                                join_type="exact", metadata_conflicts="silent")
                                                

            # plot the max power in each pixel in the same range as the right panel
            im = axes[1].imshow(pixel_max_power, extent=list(axes[0].get_xlim()) + list(axes[0].get_ylim()),
                                origin='lower', cmap='Greys', vmax=max(self.periodogram))

            cbar = fig.colorbar(im, ax=axes[1])
            cbar.set_label('LS periodogram power')
            axes[1].set_title('Maximum Lomb Scargle Power')

            axes[1].annotate((f'Frequency: {(lower + upper) / 2:1.2f} 1/day\n'
                              f'Range: [{lower:1.2f}, {upper:1.2f}] 1/day'), xy=(0.5, 0.95),
                             xycoords="axes fraction", ha="center", va="top")

            circle = plt.Circle(xy=(axes[0].get_xlim()[0] + x_range / 2, axes[0].get_ylim()[0] + y_range / 2),
                                radius=(self.radius * u.deg / TESS_RESOLUTION).to(u.pixel).value,
                                edgecolor="red", facecolor="none", linestyle="dotted")
            axes[1].add_artist(circle)

            # plot the LS periodogram for the ensemble cluster LC
            fig, axes[2] = self.plot_periodogram(self.periodogram_freqs, fig=fig, ax=axes[2], show=False,
                                                 title="Ensemble Light Curve Periodogram")
            axes[2].axvspan(lower, upper, color="lightgrey", zorder=-1)

            # save and close the figure and move on to the next
            fig.savefig(os.path.join(output_path, 'diagnostics', f'{identifier}_gif_plot_frame_{i}.png'),
                        bbox_inches="tight")
            plt.close(fig)
            i += 1

        simbad_results["MAIN_ID"] = simbad_results["MAIN_ID"].astype('str')
        simbad_results["OTYPE"] = simbad_results["OTYPE"].astype('str')

        # save the simbad results
        simbad_results.write(os.path.join(output_path, 'diagnostics',
                                          f'{identifier}_simbad_results.hdf5'),
                             path='data', serialize_meta=True, overwrite=True)


        # convert individual frames to a GIF
        gif_path = os.path.join(output_path, 'diagnostics', f'{identifier}_pixel_power_gif.gif')
        with imageio.get_writer(gif_path, mode='I', fps=2.5) as writer:
            for i in range(len(bounds)):
                writer.append_data(imageio.imread(os.path.join(output_path, 'diagnostics',
                                                               f'{identifier}_gif_plot_frame_{i}.png')))
                
        #get GIF back
        gif = HTML(f'<img src="{gif_path}">')

        return gif, simbad_results 



import matplotlib.pyplot as plt
import numpy as np


__all__ = ["plot_periodogram", "plot_acf", "plot_lightcurve"]


def plot_periodogram(frequencies, power, power_percentiles, peak_freqs, fap=None,
                     fig=None, ax=None, show=True, title=None, save_path=None):
    """Plot a periodogram

    Parameters
    ----------
    frequencies : :class:`~numpy.ndarray`
        Frequencies at which periodogram evaluated
    power : :class:`~numpy.ndarray`
        Power at each frequency
    power_percentiles : :class:`~numpy.ndarray`
        Percentiles on the power, if shape is (2, len(power)) then assumed to be 1-sigma values, if shape is
        (4, len(power)) then assumed to be 1- and 2-sigma values in order of (2-, 1-, 1+, 2+)
    peak_freqs : :class:`~numpy.ndarray`
        Frequencies at which peaks occur
    fap : `float`, optional
        False alarm probability for the maximum power peak - plotted as horizontal line if provided, by default None
    fig : :class:`~matplotlib.pyplot.Figure`, optional
        Figure on which to plot, if either `fig` or `ax` is None then new ones are created, by default None
    ax : :class:`~matplotlib.pyplot.AxesSubplot`, optional
        Axis on which to plot, if either `fig` or `ax` is None then new ones are created, by default None
    show : bool, optional
        Whether to show the plot, by default True
    title : `str`, optional
        A title for the plot, by default None
    save_path : `str`, optional
        Where to save the plot, if None then plot is not saved, by default None

    Returns
    -------
    fig, ax : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
        Figure and axis on which the periodogram has been plotted
    """
    # create figure if necessary and add title
    if fig is None or ax is None:
        fig, ax = plt.subplots()
    ax.set_title(title)

    # shade any percentiles as desired
    if power_percentiles is not None:
        if len(power_percentiles) == 2:
            one_below, one_above = power_percentiles
            ax.fill_between(frequencies, one_below, one_above, color="grey", alpha=0.3)
        else:
            two_below, one_below, one_above, two_above = power_percentiles
            ax.fill_between(frequencies, two_below, two_above, color="grey", alpha=0.15)
            ax.fill_between(frequencies, one_below, one_above, color="grey", alpha=0.3)

    # plot the actual power
    ax.plot(frequencies, power, color="black", linewidth=1.25)

    # add vertical lines at each of the peaks
    ax.vlines(x=peak_freqs, ymin=np.zeros(len(peak_freqs)),
              ymax=power[np.in1d(frequencies, peak_freqs)], linestyle='--', color="tab:red")

    if fap is not None:
        ax.axhline(fap, color="gold", label=f"False alarm probability ({fap:1.2e})")

    # set scales and labels
    ax.set_xscale('log')
    ax.set_xlabel(r'Frequency $[1 / {\rm days}]$')
    ax.set_ylabel('Power')
    ax.legend()

    # save and show if desired
    if save_path is not None:
        plt.savefig(save_path, format='png', bbox_inches="tight")
    if show:
        plt.show()

    return fig, ax


def plot_acf(time, acf, acf_percentiles=None, title=None, fig=None, ax=None, show=True, save_path=None):
    """Plot an autocorrelation function

    Parameters
    ----------
    time : :class:`~numpy.ndarray`
        Times at which autocorrelation function is evaluated
    acf : :class:`~numpy.ndarray`, optional
        The autocorrelation function, by default None
    acf_percentiles : :class:`~numpy.ndarray`, optional
        1-sigma confidence interval on the autocorrelation function, by default None
    title : `str`, optional
        A title for the plot, by default None
    fig : :class:`~matplotlib.pyplot.Figure`, optional
        Figure on which to plot, if either `fig` or `ax` is None then new ones are created, by default None
    ax : :class:`~matplotlib.pyplot.AxesSubplot`, optional
        Axis on which to plot, if either `fig` or `ax` is None then new ones are created, by default None
    show : bool, optional
        Whether to show the plot, by default True
    save_path : `str`, optional
        Where to save the plot, if None then plot is not saved, by default None

    Returns
    -------
    fig, ax : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
        Figure and axis on which the periodogram has been plotted
    """
    # create figure if necessary and add title
    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=(10, 5))
    ax.set_title(title)

    # shade percentiles if desired
    if acf_percentiles is not None:
        ax.fill_between(time, acf_percentiles[:, 0], acf_percentiles[:, 1], color='grey', alpha=.5)

    # plot autocorrelation function and label axes
    ax.plot(time, acf)
    ax.set_xlabel(r'Time $[\rm days]$')
    ax.set_ylabel("Autocorrelation")

    # save and show as desired
    if save_path is not None:
        plt.savefig(save_path, format='png', bbox_inches="tight")
    if show:
        plt.show()

    return fig, ax


def plot_lightcurve(time, flux, fold_period=None, title=None, fig=None, ax=None, show=True, save_path=None,
                    color='k', linewidth=0.5, **kwargs):
    """Plot a lightcurve

    Parameters
    ----------
    time : :class:`~numpy.ndarray`
        Times of observations
    flux : :class:`~numpy.ndarray`
        Flux of each observation
    fold_period : `float`, optional
        Period on which to fold the light curve, if None then no folding is done, by default None
    title : `str`, optional
        A title for the plot, by default None
    fig : :class:`~matplotlib.pyplot.Figure`, optional
        Figure on which to plot, if either `fig` or `ax` is None then new ones are created, by default None
    ax : :class:`~matplotlib.pyplot.AxesSubplot`, optional
        Axis on which to plot, if either `fig` or `ax` is None then new ones are created, by default None
    show : bool, optional
        Whether to show the plot, by default True
    save_path : `str`, optional
        Where to save the plot, if None then plot is not saved, by default None
    color : `str`, optional
        Colour to use for the lightcurve line, by default 'k'
    linewidth : `float`, optional
        Linewidth for the lightcurve, by default 0.5
    **kwargs
        Keyword arguments to pass to the plotting of the lightcurve line

    Returns
    -------
    fig, ax : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
        Figure and axis on which the lightcurve has been plotted
    """
    # create figure if necessary and add title
    if fig is None or ax is None:
        fig, ax = plt.subplots()
    ax.set_title(title)

    # if a period is given then fold based on that period and re-sort
    if fold_period is not None:
        time = ((time - time[0]) %fold_period)/fold_period
        order = np.argsort(time)
        time = time[order]
        flux = flux[order]

    ax.plot(time, flux, color=color, linewidth=linewidth, **kwargs)

    ax.set_xlabel(r'Time $\rm [days]$')
    ax.set_ylabel(r'Flux $[e / {\rm s}]$')

    # save and show as desired
    if save_path is not None:
        plt.savefig(save_path, format='png', bbox_inches="tight")
    if show:
        plt.show()

    return fig, ax



import numpy as np
import scipy
from statsmodels.tsa.stattools import acf as calc_acf

from scipy.signal import find_peaks, peak_widths, argrelextrema
import scipy.linalg

from astropy.table import Table
from astropy.timeseries import LombScargle

from .plot import plot_acf

__all__ = ["get_MAD", "get_sigmaG", "get_range", "get_skewness", "von_neumann_ratio", "J_stetson",
           "periodogram", "longest_contiguous_chunk", "autocorr"]


def get_MAD(flux):
    """Get the median absolute deviation of an array of fluxes

    Parameters
    ----------
    flux : :class:`~numpy.ndarray`
        Array of fluxes

    Returns
    -------
    mad : `float`
        Median absolute deviation
    """
    return np.median(np.abs(flux - np.median(flux)))

def get_sigmaG(flux):
    """Compute a rank-based estimate of the standard deviation

    Follows https://www.astroml.org/modules/generated/astroML.stats.sigmaG.html

    Parameters
    ----------
    flux : :class:`~numpy.ndarray`
        Array of fluxes

    Returns
    -------
    sigmaG : `float`
        sigmaG (a robust estimate of the standard deviation sigma)
    """
    # This factor is ~ 1 / (2 sqrt(2) erf^-1(0.5))
    sigmaG_factor = 0.74130110925280102
    percs = np.percentile(flux, [25, 75])
    return sigmaG_factor * (percs[1] - percs[0])

def get_range(flux, bottom_percentile, upper_percentile):
    # TODO: Ask Tobin but it seems like this will only work for an array of fluxes of length 100?
    # Generalized to get the range for whatever percentile the heart desires
    flux_l = list(flux)
    flux_l.sort()

    bottom_ind = int(len(flux_l)*bottom_percentile)
    upper_ind = int(len(flux_l)*upper_percentile)

    new = flux_l[bottom_ind:upper_ind]
    r = new[-1] - new[0]
    return r


def get_skewness(flux):
    """Calculate the skewness of an array of fluxes

    Parameters
    ----------
    flux : :class:`~numpy.ndarray`
        Array of fluxes

    Returns
    -------
    skewness : `float`
        Skewness
    """
    return scipy.stats.moment(flux, moment=3)


def von_neumann_ratio(flux):
    """Calculate the (inverted) Von Neumann Ratio for an array of fluxes

    See `von Neumann 1941 <https://www.jstor.org/stable/2235951>`_ and
    `Sokolovsky+2017 <https://ui.adsabs.harvard.edu/abs/2017MNRAS.464..274S/abstract>`_ Eq. 19

    Parameters
    ----------
    flux : :class:`~numpy.ndarray`
        Array of fluxes

    Returns
    -------
    ratio : `float`
        Von Neumann Ratio
    """
    eta = np.sum(np.ediff1d(flux)**2) / np.var(flux)
    return 1 / eta


def J_stetson(time, mag, mag_err, pair_threshold=0.021):
    """Calculate the J Stetson statistic as a measure of variability

    See `Stetson+1996 <https://ui.adsabs.harvard.edu/abs/1996PASP..108..851S/abstract>`_ and Eq. 12 of
    `Sokolovsky+2017 <https://ui.adsabs.harvard.edu/abs/2017MNRAS.464..274S/abstract>`_

    Parameters
    ----------
    time : :class:`~numpy.ndarray`
        Time of observations
    mag : :class:`~numpy.ndarray`
        Magnitudes of each observation
    mag_err : :class:`~numpy.ndarray`
        Errors on the magnitudes
    pair_threshold : `float`, optional
        Threshold for whether to consider observations close enough, by default 0.021 days (=~30 min)

    Returns
    -------
    J : `float`
        J Stetson Statistic
    """
    # get the difference in time between observations
    delta_times = np.concatenate((np.ediff1d(time), [1]))

    # calculate the delta for each observation
    n = len(time)
    delta = (np.sqrt(n / (n - 1)) * ((mag - np.mean(mag)) / mag_err))

    # start the P and weights
    P = np.empty(len(delta_times))
    w = np.ones(len(delta_times))

    # mask for which observations are part of a pair (0.021 day=~30min)
    pairs = delta_times < pair_threshold
    inds = np.argwhere(pairs).flatten()

    # calculate P and w
    P[inds] = delta[inds] * delta[inds]
    P[~pairs] = delta[~pairs]**2 - 1
    w[~pairs] = 0.25

    # combine into the J stetson stat
    J = np.sum(w * np.sign(P) * np.sqrt(np.abs(P))) / np.sum(w)
    return J


def periodogram(time, flux, flux_err, frequencies, n_bootstrap=100, max_peaks=25, freq_thresh=5):
    """Generate a bootstrapped Lomb-Scargle periodogram and various associated statistics

    Parameters
    ----------
    time : :class:`~numpy.ndarray`
        Times of observations
    flux : :class:`~numpy.ndarray`
        Flux for each observations
    flux_err : :class:`~numpy.ndarray`
        Errors on the fluxes
    frequencies : :class:`~numpy.ndarray`
        Frequencies at which to evaluate the periodogram
    n_bootstrap : `int`, optional
        How many bootstraps to perform, by default 100
    max_peaks : `int`, optional
        Fixed number of max peaks to return, by default 25. If None is given then the number of peaks that
        exist will be returned rather than truncated/padded to this fixed number.
    freq_thresh : `int`, optional
        Threshold for high vs. low frequency, by default 5 days since this represents a threshold of how
        different stars vary
        `McQuillan+2012 <https://ui.adsabs.harvard.edu/abs/2013ApJ...775L..11M/abstract>`_.

    Returns
    -------
    med : :class:`~numpy.ndarray`
        Median power at each frequency
    percentiles : :class:`~numpy.ndarray`
        1 and 2-sigma power at each frequency (in order of 2-, 1-, 1+, 2+) with shape `(4, len(frequencies))`
    lsp_stats : `dict`
        Dictionary of various statistics related to the periodogram. Keys are: ["max_power",
        "freq_at_max_power", "peak_freqs", "power_at_peaks", "n_peaks", "ratio_of_power_at_high_v_low_freq"].
        Many of these are self explanatory but we note that `peak_freqs` and `power_at_peaks` will always have
        a length of `max_peaks` may need to be masked to just take the first `n_peaks`. (This feature is
        useful when saving to a FITS file with a required fixed column width).
    """
    flux_samples = np.transpose([np.repeat(flux[i], n_bootstrap)
                                 + np.random.normal(loc=0,
                                                    scale=flux_err[i],
                                                    size=n_bootstrap) for i in range(len(flux))])
    powers = np.asarray([LombScargle(time, flux_samples[i], dy=flux_err).power(frequencies)
                         for i in range(len(flux_samples))])

    percentiles = np.percentile(powers, [50, 2.27, 15.89, 84.1, 97.725], axis=0)
    med, _, one_below, _, _ = percentiles

    peak_threshold = np.mean(med) + 2 * np.std(med)
    peak_inds, peak_info = find_peaks(one_below, height=peak_threshold)

    _, _, peak_left, peak_right = peak_widths(one_below, peak_inds, rel_height=0.5)
    peak_left = frequencies[np.floor(peak_left).astype(int)]
    peak_right = frequencies[np.ceil(peak_right).astype(int)]

    # calculate how much of our power is coming from timescales shorter/longer than threshold
    low_freq = frequencies > freq_thresh
    ratio_of_power_at_high_v_low_freq = np.mean(med[~low_freq]) / np.mean(med[low_freq])

    # if no limit on the peaks then save all of them, otherwise truncate to ensure identical list length
    n_peaks = len(peak_inds)
    max_peaks = n_peaks if max_peaks is None else max_peaks

    peak_freqs, power_at_peaks = np.zeros(max_peaks), np.zeros(max_peaks)

    peak_freqs[:min(n_peaks, max_peaks)] = frequencies[peak_inds][:max_peaks]
    power_at_peaks[:min(n_peaks, max_peaks)] = peak_info['peak_heights'][:max_peaks]

    max_power = max(med)
    freq_at_max_power = frequencies[med == max_power][0]

    ls = LombScargle(time, flux, flux_err)
    ls.power(frequencies)
    fap = ls.false_alarm_probability(power=max_power)

    lsp_stats = {
        "max_power": max_power,
        "freq_at_max_power": freq_at_max_power,
        "peak_freqs": peak_freqs,
        "peak_left_edge": peak_left,
        "peak_right_edge": peak_right,
        "power_at_peaks": power_at_peaks,
        "n_peaks": n_peaks,
        "ratio_of_power_at_high_v_low_freq": ratio_of_power_at_high_v_low_freq,
        "FAP": fap
    }

    return med, percentiles[1:], lsp_stats


def longest_contiguous_chunk(times, largest_gap_allowed=0.25):
    """Create a mask for the largest contiguous chunk of observations

    Parameters
    ----------
    times : :class:`~numpy.ndarray`
        Array of times at which observations were taken
    largest_gap_allowed : `float`, optional
        Largest gap in the data which is allowed in a chunk (in days), by default 0.25

    Returns
    -------
    chunk_mask : :class:`~numpy.ndarray`
        Boolean mask on observations in the largest chunk
    """
    # find the difference in time between observations
    delta_times = np.ediff1d(times)

    # identify inds at which the change is larger than the maximum allowed and append start/end
    gap_inds = np.argwhere(delta_times > largest_gap_allowed).flatten()
    gap_inds = np.concatenate(([-1], gap_inds, [len(times) - 1]))

    # find the largest length of time between gaps
    max_ind = np.argmax(np.ediff1d(gap_inds))

    # create a mask for that longest chunk
    all_inds = np.arange(len(times))
    chunk_mask = (all_inds > gap_inds[max_ind]) & (all_inds <= gap_inds[max_ind + 1])
    return chunk_mask


def autocorr(time, flux, largest_gap_allowed=0.25, plot=False, **plot_kwargs):
    """Calculate the autocorrelation function for an array of fluxes

    Parameters
    ----------
    time : :class:`~numpy.ndarray`
        Times of observations
    flux : :class:`~numpy.ndarray`
        Flux for each observations
    largest_gap_allowed : `float`, optional
        Largest gap in the data which is allowed in a chunk (in days), by default 0.25
    plot : `bool`, optional
        Whether to show a plot of the autocorrelation, by default False
    plot_kwargs : `various`
        Arguments to pass to `elk.plot.plot_acf`

    Returns
    -------
    ac_time : :class:`~numpy.ndarray`
        Times evaluated for the autocorrelation function
    acf : :class:`~numpy.ndarray`
        The autocorrelation function
    confint : :class:`~numpy.ndarray`
        1-sigma confidence interval around the autocorrelation function with shape (len(ac_time), 2)
    acf_stats : `dict`
        Some simple stats for the autocorrelation function
    fig, ax : :class:`~matplotlib.pyplot.Figure`, :class:`~matplotlib.pyplot.AxesSubplot`
        Figure and axis on which the correlation function has been plotted
    """
    chunk_mask = longest_contiguous_chunk(time, largest_gap_allowed=largest_gap_allowed)

    ac_time, ac_flux = time[chunk_mask], flux[chunk_mask]
    acf, confint = calc_acf(ac_flux, nlags=len(ac_flux) - 1, alpha=0.3173)
    plot_times = ac_time - ac_time[0]

    # find max autocorrelation (cut first value since it is meaningless in this context)
    max_ac = max(acf[plot_times > plot_times[argrelextrema(acf, np.less)[0][0]]])
    acf_stats = {
        "max_autocorrelation": max_ac,
        "time_of_max_autocorrelation": plot_times[acf == max_ac],
        "rms": np.sqrt(np.mean(np.square(acf)))
    }

    if plot:
        fig, ax = plot_acf(plot_times, acf, confint, **plot_kwargs)
        return ac_time, acf, confint, acf_stats, fig, ax
    else:
        return ac_time, acf, confint, acf_stats
